#!/usr/bin/python3

# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# cin_seff --- https://gitlab.hpc.cineca.it/ddibari0/cin_seff
# 
# Released under the GNU Public Licence, v2 or any higher version
#

"""\
Report the efficiency of jobs on CINECA Clusters
===================================================================

.. versionadded:: 0.1

TODO: Add in-depth discussion.

Helper functions
----------------
.. autofunction:: run
.. autofunction:: strclean
.. autofunction:: get_args
"""

import subprocess
import argparse
import sys

def run(cmd, timeout=60, splitted_cmd=False):
    """ 
    Wrapper of subprocess.run used to run the command described by cmd_list, and get back the results on the stdout and stderr. It waits for command to complete, then return a CompletedProcess instance.
    
    Parameters
    ----------
    cmd : str or list
        Command to run.
        
    timeout : int
        Timeout in seconds. After this time, if not responding, the funtion will throw an error (TimeoutExpired). 
        Default: 60
        
    splitted_cmd : bool
        The cmd is already splitted? 
        Default: False

    Returns
    -------
    proc : subprocess.CompletedProcess
        The return value from run(), representing a process that has finished.
    """
    if not splitted_cmd:
        cmd = cmd.split()
    try:
        # Run all commands and get the outputs
        return subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='UTF-8', timeout=timeout)
    except subprocess.TimeoutExpired as timeout_error:
        raise timeout_error

def strclean(s, crms=["\""," "], case="lower", strip=True, rm_all_wspaces=False):
    """ Clean string from specific characters

    Parameters
    ----------
    s : string
        Input string to clean.
    strip : bool, optional
        Strip the string (i.e. remove whitespaces from the begin and the end of the input string). 
        Default: True.
    rm_all_wspaces : bool, optional
        Remove all whitespaces from the string. 
        Default: False.
    cmrs : list, optional 
        Characters to remove from the input string. 
        Default: ["\"", " "].  
    case : string, optional
        Select if return the string with all lowecase characters (case='lower'), all uppercase characters (case='upper'), the first character of the string to an uppercase letter and all other alphabets to lowercase, or initial character in each word to uppercase and the subsequent characters to lowercase (case='title'). 
        Default: 'lower'.

    Returns
    -------
    s : string
        Cleaned string.
    """
    
    if strip:
        s = s.strip()

    # Remove specific characters
    _wspaces = (" ", "\t", "\v", "\n", "\r", "\f")
    if rm_all_wspaces:
        crms = _wspaces
    for c in crms:
        if c in s:
            s = s.replace(c,"")

    # Set case of string characters
    if case != "":
        case = case.lower()
        if case == "lower":
            s = s.lower()
        elif case == "upper":
            s = s.upper()
        elif case == "capitalize":
            s.capitalize()
        elif case == "title":
            s.title()
        else:
            print(f"\nWARNING! Unknown parameter: '{case}'. Allowed params are: 'lower', 'upper', 'capitalize', 'title'.")
            sys.exit()

    return s

def get_args():
    """ Parse Input Arguments 

    This program provides statistics related to the efficiency of resource usage by the completed job.
    
    Input options:
    --------------
    -h, --help            
        show this help message and exit
    -j JOBID, --jobid JOBID                    
        ID of the completed job. 
    -c CONSUMED_BUDGET, --consumed_budget CONSUMED_BUDGET
        Report the budget consumed by the job.
        Default: True
    -d NODE_DISTRIBUTION, --node_distribution NODE_DISTRIBUTION
        Report the distribution of the used compute nodes over the network.
        Default: False
    -v VERBOSITY, --verbosity VERBOSITY
        Set the verbosity level.
        Default: 0
    
    Returns:
    --------
    args : argparse.ArgumentParser
        Instantied object of the class argparse.ArgumentParser where all argument strings were
        converted to objects and assign them as attributes of the namespace. Return the populated namespace.
    """
    
    parser = argparse.ArgumentParser(prog="CheckHL", description="This program allows you to verify the last time that a user logged on a CINECA cluster.")
    
    parser.add_argument(
        "-j", "--jobid", 
        help="ID of the completed job."
    )

    parser.add_argument(
        "-c", "--consumed_budget", 
        action = "store_false",
        help = "Report the budget consumed by the job."
    )

    parser.add_argument(
        "-d", "--node_distribution", 
        action = "store_true",
        help = "Report the distribution of the used compute nodes over the network."
    )

    parser.add_argument(
        "-v", "--verbosity",
        type = int, 
        default = 0,
        help = "Set program verbosity. Used for debugging and testing of the application."
    )
    
    return parser.parse_args()

def from_sec_to_ft(time_in_sec):

    secs = float(time_in_sec)
    
    days = int(secs / 86400)
    residue = secs - days*86400
    hours   = int(residue / 3600) 
    residue = residue - hours*3600
    minutes = int(residue / 60)
    seconds = residue - minutes*60

    out=""
    if days != 0:
        out=f"{days}-" 
    out+=f"{hours}:{minutes}:{seconds}"

    return out
    
def from_ft_to_sec(ft):
    time_in_seconds = 0

    idx_sep_days = ft.find('-')
    if idx_sep_days != -1:
        # Time format == D-HH:MM:SS
        days=float(ft[:idx_sep_days])
        hhmmss = ft[idx_sep_days+1:].split(':')
        hh = float(hhmmss[0])
        mm = float(hhmmss[1])
        ss = float(hhmmss[2])
        time_in_seconds = days*86400 + hh*3600 + mm*60 + ss
    else:
        # Time format == HH:MM:SS or MM:SS
        tmp = ft.split(':')

        if len(tmp) == 3:
            # Time format == HH:MM:SS
            hh = float(tmp[0])
            mm = float(tmp[1])
            ss = float(tmp[2])
            time_in_seconds = hh*3600 + mm*60 + ss
        else:
            # Time format == MM:SS
            mm = float(tmp[0])
            ss = float(tmp[1])
            time_in_seconds = mm*60 + ss
            
    return time_in_seconds

def main():
    args = get_args()
    
    if args.verbosity > 2:
        print(f"PARSED VARIABLES:\njobid=={args.jobid}\nconsumed_budget=={args.consumed_budget}\nnode_distribution=={args.node_distribution}\nverbosity=={args.verbosity}")

    cmd = "sacct -P -n -a --format JobID,User,Account,State,AllocCPUS,REQMEM,TotalCPU,ElapsedRaw,MaxRSS,ExitCode,NodeList,AllocNodes,Elapsed -j {jobid}"
    proc = run(cmd.format(jobid=args.jobid))
    if proc.stderr != "":
        raise Exception(f"Command:\n{cmd}\nERROR:\n{proc.stderr}")
    if args.verbosity > 2:
        print(f"\nCommand:\n{cmd.format(jobid=args.jobid)}\nOUTPUT:\n{proc.stdout}")
        
    rawdata = proc.stdout.splitlines()
    
    data = dict(TotalRSS=0.0)
    
    main_info = rawdata[0].split('|')
    
    data["JobID"]       = args.jobid
    data["User"]        = main_info[1]
    data["Account"]     = main_info[2]
    data["State"]       = main_info[3]
    data["AllocCPUS"]   = int(main_info[4])
    data["REQMEM"]      = main_info[5]
    data["TotalCPU"]    = main_info[6]
    data["ElapsedRaw"]  = float(main_info[7])
    data["ExitCode"]    = main_info[9]
    data["NodeList"]    = main_info[10]
    data["AllocNodes"]  = int(main_info[11])
    data["Elapsed"]     = main_info[12]
    data["CPUSperNode"] = data["AllocCPUS"] / data["AllocNodes"]
    data["CoreWallTimeRaw"] = data["ElapsedRaw"] * data["AllocCPUS"]
    data["CoreWallTime"]    = from_sec_to_ft(data["CoreWallTimeRaw"])
    data["TotalCPURaw"] = from_ft_to_sec(data["TotalCPU"])
    data["CPUEfficiency"]   = 100 * data["TotalCPURaw"] / data["CoreWallTimeRaw"]
    
    for line in rawdata[1:]:
        step_info = line.split('|')
        
        # Get unit from TotalRSS
        unit=step_info[8][-1]
        
        # Raise an error if different from kiB
        if unit != "K":
            raise Exception(f"ERROR: unit of MaxRSS ({unit}) unrecongnized")
        
        step_MAXRSS = float(step_info[8][:-1])
        if step_MAXRSS > data["TotalRSS"]:
            data["TotalRSS"] = step_MAXRSS

    data["TotalRSS_GiB"] = data["TotalRSS"] / 1048576
    
    # Get unit from REQMEM
    unit=data["REQMEM"][-1]
    # Convert REQMEM to kiB
    if unit == "T":
        # from GiB to kiB (1024**2==1048576)
        data["REQMEM_kiB"] = float(data["REQMEM"][:-1]) * 1073741824
    elif unit == "G":
        # from GiB to kiB (1024**2==1048576)
        data["REQMEM_kiB"] = float(data["REQMEM"][:-1]) * 1048576
    elif unit == "M":
        # from GiB to kiB (1024**2==1048576)
        data["REQMEM_kiB"] = float(data["REQMEM"][:-1]) * 1024 
    elif unit == "K":
        # from GiB to kiB (1024**2==1048576)
        data["REQMEM_kiB"] = float(data["REQMEM"][:-1])
    else:
        raise Exception(f"ERROR: unit of REQMEM ({unit}) unrecongnized")
    
    data["REQMEM_GiB"] = data["REQMEM_kiB"] / 1048576
    data["MEMEfficiency"] = 100 * data["TotalRSS"] / data["REQMEM_kiB"]

    report = f"""
Job ID:  {data["JobID"]}
User:    {data["User"]}
Account: {data["Account"]}
State: {data["State"]} (exit code {data["ExitCode"].split(":")[0]})
Nodes: {data["AllocNodes"]}
Cores per node: {data["CPUSperNode"]}
CPU Utilized: {data["TotalCPU"]}
CPU Efficiency: {data["CPUEfficiency"]:.2f}% of {data["CoreWallTime"]} core-walltime
Job Wall-clock time: {data["Elapsed"]}
Memory Utilized: {data["TotalRSS_GiB"]} GiB
Memory Efficiency: {data["MEMEfficiency"]:.2f}% of {data["REQMEM_GiB"]} GiB
"""
    print(report)

if __name__ == "__main__":
    main()
