#!/usr/bin/python3

# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# cin_seff --- https://gitlab.hpc.cineca.it/ddibari0/cin_seff
# 
# Released under the GNU Public Licence, v2 or any higher version
#

"""\
strman: STRing MANipulations
===================================================================

.. versionadded:: 0.1

Libraries of functions used to manipulate stings, exanding 
characters or cleaning them from udesiderate patherns.

Helper functions
----------------
.. autofunction:: strclean
.. autofunction:: xpnd_lrange
.. autofunction:: xpnd_first_str_range
.. autofunction:: xpnd_full_ranges
"""

# String Cleaning

def clean(s, strip=True, rm_all_wspaces=False, rplc={}, crms=["\""," "], case="lower"):
    """ Clean string from specific characters

    Parameters
    ----------
    s : string
        Input string to clean.
    strip : bool, optional
        Strip the string (i.e. remove whitespaces from the begin and the end of the input string). 
        Default: True.
    rm_all_wspaces : bool, optional
        Remove all whitespaces from the string. 
        Default: False.
    strip : dict, optional 
        Pairs of characters to replace. Syntax: {old_char: new_char}. 
        Default: {}.
    cmrs : list, optional 
        Characters to remove from the input string. 
        Default: ["\"", " "].  
    case : string, optional
        Select if return the string with all lowecase characters (case='lower'), all uppercase characters (case='upper'), the first character of the string to an uppercase letter and all other alphabets to lowercase, or initial character in each word to uppercase and the subsequent characters to lowercase (case='title'). 
        Default: 'lower'.

    Returns
    -------
    s : string
        Cleaned string.
    """
    
    def remove_chars(sin, chars):
        for c in chars:
            if c in sin:
                sin = sin.replace(c,"")
        return sin
    
    if strip:
        s = s.strip()
            
    # Remove all withespaces
    _wspaces = (" ", "\t", "\v", "\n", "\r", "\f")
    if rm_all_wspaces:
        s = remove_chars(s, _wspaces)

    # Replace specific characters pairs
    for c_old, c_new in rplc.items():
        s = s.replace(c_old, c_new)
            
    # Remove specific characters
    s = remove_chars(s, crms)        
            
    # Set case of string characters
    if case != "":
        case = case.lower()
        if case == "lower":
            s = s.lower()
        elif case == "upper":
            s = s.upper()
        elif case == "capitalize":
            s.capitalize()
        elif case == "title":
            s.title()
        else:
            raise Exception(f"\nWARNING! Unknown parameter: '{case}'. Allowed params are: 'lower', 'upper', 'capitalize', 'title'.")

    return s

## Expand slurm lists

def xpnd_lrange(snums, unsafe=False, verbose=True):
    num_ranges = snums.split(",")
    
    if len(num_ranges) > 0:
        num_list = []
        for nr in num_ranges:
            if "-" in nr:
                limits = nr.split("-")
                if len(limits) != 2:
                    raise Exception(f"Problem with range: {nr}")
                n_digits = len(limits[0])
                if len(limits[1]) != n_digits:
                    raise Exception(f"Different number of digits for range limits: {limits[0]} -> {limits[1]}")
                start = int(limits[0])
                end   = int(limits[1])
                for i in range(start, end+1):
                    num_list.append("{0:0{1}d}".format(i,n_digits))
            else:
                num_list.append(nr)
        return num_list
    else:
        out_message = f"No range to expand found in '{snums}'"
        if unsafe:
            if verbose:
                print(f"WARNING! {out_message}. The string will be used anyway.")
            return [int(snums)]  
        else:
            raise Exception(out_message)      

def xpnd_first_str_range(srange, unsafe=False, verbose=True):
    if srange.find("[") > 0:
        limits = [srange.find("["), srange.find("]")]
        prefix = srange[:limits[0]]
        suffix = srange[limits[1]+1:]
        snums  = srange[limits[0]+1:limits[1]]
        lnums  = xpnd_lrange(snums)
        lrange = []
        for n in lnums:
            lrange.append(prefix + n + suffix)
        return lrange
    else:
        out_message = f"No range to expand found in '{srange}'"
        if unsafe:
            if verbose:
                print(f"WARNING! {out_message}. The string will be used anyway.")
            return [srange]
        else:
            raise Exception(out_message)
    
def xpnd_full_ranges(sranges, unsafe=False, verbose=True):
    n_of_expansions = sranges.count("[")
    if n_of_expansions != sranges.count("]"):
        raise Exception(f"Different number of '[' and ']' in: {sranges}")
    
    if n_of_expansions > 0:
        lranges = xpnd_first_str_range(sranges)
        for i in range(n_of_expansions-1):
            tmp = []
            for srange in lranges:
                print(srange)
                tmp.extend(xpnd_first_str_range(srange))
            lranges = tmp
        return lranges
    else:
        out_message = f"No range to expand found in '{sranges}'"
        if unsafe:
            if verbose:
                print(f"WARNING! {out_message}. The string will be used anyway.")
            return [sranges]
        else:
            raise Exception(out_message)
